/* Tunguska, ternary virtual machine
 *
 * Copyright (C) 2007,2008 Viktor Lofgren
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "machine.h"
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>

typedef enum { STR = 1, NUM, SYM, LBL } typet;
typedef enum { DOLLAR = 5, INDIR, AND, HASH, AT, COMMA, LPARAN, RPARAN, X, Y, A, PLUS } symt;
typedef struct { 
		typet type; 
		union {
			char* str;
			int num;
			symt sym;
		} data; 
} token; 
#define END -1

/* 100 tokens per line ought to be enough for anyone ;) */
token tokens[100];
int tokenn;

/* Quell lookup debug messages when doing initial sweep */
int shutup = 0; 

/* Chatty mode */
int verbose = 0;

void add_sym(symt sym) {
	tokens[tokenn].type = SYM;
	tokens[tokenn].data.sym = sym;
	tokenn++;
}

void add_num(int num) {
	tokens[tokenn].type = NUM;
	tokens[tokenn].data.num = num;
	tokenn++;
}

void add_str(char* s) {
	tokens[tokenn].type = STR;
	tokens[tokenn].data.str = strdup(s);
	tokenn++;
}
void add_lbl(char* s) {
	tokens[tokenn].type = LBL;
	tokens[tokenn].data.str = strdup(s);
	tokenn++;
}

void tokenize(char* string) {
	tokenn = 0;
	char* substring = string + strspn(string, " \t\n");
	while(substring[0] != ';' && strcspn(substring, " \t\n")) {
		if(substring[0] == '\'') {
			char* s = strndup(substring+1, strcspn(substring+1, "\n\'"));
			if(strlen(s)) add_str(s);
			substring = substring+1 + strcspn(substring+1, "\n\'");
			substring++;
			continue;
		}
		char* s = strndup(substring, strcspn(substring, " \t\n\'"));
		char* sp = s; 
		
		if(substring[0] == ';') break;
symbol:
		if(*sp == '>') { add_sym(INDIR); sp++; goto symbol; }
		else if(*sp == ',') { add_sym(COMMA); sp++; goto symbol; }
		else if(*sp == '#') { add_sym(HASH); sp++; goto symbol; }
		else if(*sp == '(') { add_sym(LPARAN); sp++; goto symbol; }
		else if(*sp == ')') { add_sym(RPARAN); sp++; goto symbol; }
		else if(*sp == '+') { add_sym(PLUS); sp++; goto symbol; }
		else if(*sp == 'X') { add_sym(X); sp++; goto symbol; }
		else if(*sp == 'Y') { add_sym(Y); sp++; goto symbol; }
		else if(*sp == 'A' && !isalpha(*(sp+1))) { add_sym(A); sp++; goto symbol; }
		else if(*sp == '$') {
			int num; int len;
			if(sscanf(sp, "$%d%n", &num, &len) <= 0) {
				printf("Invalid numeral!");
				num = 0;
			} 
			add_num(num);
			sp = sp + len; goto symbol;
		} else if(*sp == '%') {
			int len = strspn(sp+1, "ABCD01234");
					       
			if(len == 3) {
				char s1[4] = { sp[1], sp[2], sp[3], 0 };
				add_num(tryte(s1).to_int()); 
			} else if(len == 6) {
				char s1[4] = { sp[1], sp[2], sp[3], 0 };
				char s2[4] = { sp[4], sp[5], sp[6], 0 };
				add_num(tryte::word_to_int(s1, s2)); 
			} else {
				printf("Malformed balanced nonary number");
			}
			sp += len+1; goto symbol;

		} else if(*sp == '@') {
			char* s2 = strndup(sp+1, strcspn(sp+1, " )+\t\n,"));
			add_lbl(s2);
			free(s2);
			sp = sp + strcspn(sp, " )+\t\n,");
			goto symbol;
		} else if(strlen(sp)) {
			char* s2 = strndup(sp, strcspn(sp, " \t\n,@"));
			if(strlen(s2)) add_str(s2);
			free(s2);
			sp = sp + strcspn(sp, " \t\n,@");
			goto symbol;
		}
		free(s);
		substring = substring + strcspn(substring, " \t\n\'");
		substring = substring + strspn(substring, " \t\n");
	}
}

#define LABELMAX 5000
typedef struct { char* label; int mempos; } labelt;
labelt labels[LABELMAX];
int labeln = 0;

void add_label(char* s, int pos) {
	int local = 0;
	if(s[0] == '.') {
		local = 1;
		int lab = labeln;
		for(lab = labeln-1; lab >= 0; lab -- ) {
			if(index(labels[lab].label, '.') == NULL) {
				int len = strlen(labels[lab].label) +
						strlen(s) + 1;
				char* s2 = (char*)
				       	malloc(len);
				snprintf(s2, len, "%s%s", labels[lab].label, s);
				s = s2;
				break;
			}
		}
		if(lab < 0) printf("Local label before global label!\n");
	}
	for(int i = 0; i < labeln; i++) 
		if(strcasecmp(labels[i].label, s) == 0) {
			 printf("Duplicate label %s!!", s);
			 return;
		}
	if((unsigned int) labeln >= sizeof(labels)/sizeof(labels[0])) {
		printf("Out of label-space!");
	}
	if(!local) labels[labeln].label = strdup(s);
	else labels[labeln].label = s;

	labels[labeln].mempos = pos;
	labeln++;
}

int lookup_label(char* s, int pc) {
	int local = 0; 
	if(s[0] == '.') {
		local = 1;
		int closest = 0;
		int closest_pc = 0;
		for(int lab = 0; lab < labeln; lab++) {
			if(labels[lab].mempos <= pc && index(labels[lab].label, '.') == 0)
		       	{
				if(pc - labels[lab].mempos < pc - closest_pc) {
					closest = lab;
					closest_pc = labels[lab].mempos;
				}
			}
		}

		int len = strlen(labels[closest].label) + strlen(s) + 1;
		char *s2 = (char*) malloc(len);
		snprintf(s2, len, "%s%s", labels[closest].label, s);
		s = s2;
	}
	for(int i = 0; i < labeln; i++) {
		if(strcmp(s,labels[i].label)==0) return labels[i].mempos;
	}
	if(!shutup) printf("Label mismatch '%s'!\n", s);
	if(local) free(s);
	return 0;
}


unsigned int strtoopc(char* s) {
	int i;
	for(i = -40; i <= 40; i++) {
		if(strcasecmp(s, machine::opcode_to_string(i)) == 0) {
			return i;
		}
	}

	return 0xdeadbeef;
}

/* Syntax matching function */
bool match_syntax(int shift, ...) {
	va_list ap; int pos = 0;
	va_start(ap, shift);
	for(;;) {
		int v = va_arg(ap, int);
		if(pos >= tokenn - shift - 1) {
			if(v == END) break;
			goto mismatch; 
		}
		if(v == NUM || v == LBL || v == STR) { if(tokens[shift+1+pos].type != v) goto mismatch; }
		else if(tokens[shift+1+pos].type != SYM ||
			   tokens[shift+1+pos].data.sym != v) {
			goto mismatch;
		}
		pos ++;
	}
	va_end(ap);
	return true;
mismatch:
	va_end(ap);
	return false;

}

// Be afraid. Be so very afraid.
//
// TODO: Re-implement with tryte::int_to_word
int addop(int opcode, int shift, machine& m, int pc) {

	if(tokenn-shift-1 == 0) { m.memref(pc) = m.qop(machine::IMPLICIT, opcode); return 1; }
	if(match_syntax(shift, A, END)) { m.memref(pc) = m.qop(machine::ACC, opcode); return 1; }  // Accumulator
	if(match_syntax(shift, HASH, NUM, END)) {
		m.memref(pc) = m.qop(machine::IMMEDIATE, opcode);  // Immediate addressing
	        m.memref(pc+1) = tokens[shift+2].data.num; 
		return 2; 
	}

	if(match_syntax(shift, LBL, END)) {			// Absolute addressing (label)
		m.memref(pc) =   m.qop(machine::ABS, opcode);
	        m.memref(pc+2) = tryte(lookup_label(tokens[shift+1].data.str, pc)); 
		m.memref(pc+1) = (lookup_label(tokens[shift+1].data.str, pc)-
				m.memref(pc+2).to_int())/729;
		return 3; 
	}

	if(match_syntax(shift, LBL, PLUS, NUM,  END)) {			// Absolute addressing (label + OFFSET)
		m.memref(pc) =   m.qop(machine::ABS, opcode);
		int maddr = lookup_label(tokens[shift+1].data.str, pc) + tokens[shift+3].data.num;
		tryte::int_to_word(maddr, m.memref(pc+1), m.memref(pc+2));
		return 3; 
	}
	
	if(match_syntax(shift, PLUS, NUM, END)) { // Branch-style jump
		int pos = pc + 3 + tokens[shift+2].data.num;

		m.memref(pc) = m.qop(machine::ABS, opcode);
	        m.memref(pc+2) = tryte(pos);
		m.memref(pc+1) = (pos-m.memref(pc+2).to_int())/729;
		return 3;
	}
	if(match_syntax(shift, NUM, END)) {	// Absolute addressing (label)
		m.memref(pc) =   m.qop(machine::ABS, opcode);
	        m.memref(pc+2) = tryte(tokens[shift+1].data.num); 
		m.memref(pc+1) = (tokens[shift+1].data.num - 
				m.memref(pc+2).to_int())/729;
		return 3; 
	}


	if(match_syntax(shift, LPARAN, LBL, RPARAN, END)) { // Indirect addressing (label)
		m.memref(pc) = m.qop(machine::INDIRECT, opcode);
	        m.memref(pc+2) = tryte(lookup_label(tokens[shift+2].data.str, pc)); 
	        m.memref(pc+1) = (lookup_label(tokens[shift+2].data.str, pc)-m.memref(pc+2).to_int())/729;
		return 3;
	}

	if(match_syntax(shift, LPARAN, NUM, RPARAN, END)) { // Indirect addressing (number)
		m.memref(pc) = m.qop(machine::INDIRECT, opcode);
	        m.memref(pc+2) = tryte(tokens[shift+2].data.num); 
	        m.memref(pc+1) = (tokens[shift+2].data.num-m.memref(pc+2).to_int())/729;
		return 3; 
	}

	if(match_syntax(shift, LPARAN, NUM, COMMA, X, RPARAN, END)) { // Indirect,X addressing (num)
		m.memref(pc) = m.qop(machine::INDX, opcode);
		m.memref(pc+2) = tokens[shift+2].data.num; 
		m.memref(pc+1) = (tokens[shift+2].data.num-m.memref(pc+2).to_int())/729;
		return 3;
	}
	if(match_syntax(shift, LPARAN, LBL, COMMA, X, RPARAN, END)) { // Indirect,X addressing (lbl)
		m.memref(pc) = m.qop(machine::INDX, opcode);
		m.memref(pc+2) = tryte(lookup_label(tokens[shift+2].data.str, pc)); 
		m.memref(pc+1) = (lookup_label(tokens[shift+2].data.str, pc)-m.memref(pc+2).to_int())/729;
		return 3;
	} 
	if(match_syntax(shift, LPARAN, LBL, PLUS, NUM, COMMA, X, RPARAN, END)) { // Indirect,Y addressing (lbl+offset)
		m.memref(pc) = m.qop(machine::INDX, opcode);
		int maddr = lookup_label(tokens[shift+2].data.str, pc) + tokens[shift+4].data.num;
		tryte::int_to_word(maddr, m.memref(pc+1), m.memref(pc+2));
		return 3;
	}

	if(match_syntax(shift, LPARAN, NUM, COMMA, Y, RPARAN, END)) { // Indirect,Y  addressing (num)
		m.memref(pc) = m.qop(machine::INDY, opcode);
		m.memref(pc+2) = tokens[shift+2].data.num; 
		m.memref(pc+1) = (tokens[shift+2].data.num-m.memref(pc+2).to_int())/729;
		return 3;
	}

	if(match_syntax(shift, LPARAN, LBL, COMMA, Y, RPARAN, END)) { // Indirect,Y addressing (lbl)
		m.memref(pc) = m.qop(machine::INDY, opcode);
		m.memref(pc+2) = tryte(lookup_label(tokens[shift+2].data.str, pc)); 
		m.memref(pc+1) = (lookup_label(tokens[shift+2].data.str, pc)-m.memref(pc+2).to_int())/729;
		return 3;
	}
	if(match_syntax(shift, LPARAN, LBL, PLUS, NUM, COMMA, Y, RPARAN, END)) { // Indirect,Y addressing (lbl+offset)
		m.memref(pc) = m.qop(machine::INDY, opcode);
		int maddr = lookup_label(tokens[shift+2].data.str, pc) + tokens[shift+4].data.num;
		tryte::int_to_word(maddr, m.memref(pc+1), m.memref(pc+2));
		return 3;
	}

	if(match_syntax(shift, NUM, COMMA, X, END)) { // Absolute,X addressing (num)
	        m.memref(pc) = m.qop(machine::AX, opcode);
		m.memref(pc+2) = tokens[shift+1].data.num; 
		m.memref(pc+1) = (tokens[shift+1].data.num-m.memref(pc+2).to_int())/729;
	       	return 3; 
	} 

	if(match_syntax(shift, LBL, COMMA, X, END)) { // Absolute,X addressing (lbl)
	        m.memref(pc) = m.qop(machine::AX, opcode);
		m.memref(pc+2) = tryte(lookup_label(tokens[shift+1].data.str, pc)); 
		m.memref(pc+1) = (lookup_label(tokens[shift+1].data.str, pc)-m.memref(pc+2).to_int())/729;
	       	return 3; 
	}	
	
	if(match_syntax(shift, LBL, PLUS, NUM, COMMA, X, END)) { // Absolute,X addressing (lbl+offset)
	        m.memref(pc) = m.qop(machine::AX, opcode);
		int maddr = lookup_label(tokens[shift+1].data.str, pc) + tokens[shift+3].data.num;
		tryte::int_to_word(maddr, m.memref(pc+1), m.memref(pc+2));
	       	return 3; 
	}

	if(match_syntax(shift, NUM, COMMA, Y, END)) { // Absolute,Y addressing (num)
		m.memref(pc) = m.qop(machine::AY, opcode);
		m.memref(pc+2) = tokens[shift+1].data.num; 
		m.memref(pc+1) = (tokens[shift+1].data.num-m.memref(pc+2).to_int())/729;
		return 3; 
	} if(match_syntax(shift, LBL, COMMA, Y, END)) { // Absolute,X addressing (lbl)
	        m.memref(pc) = m.qop(machine::AY, opcode);
		m.memref(pc+2) = tryte(lookup_label(tokens[shift+1].data.str, pc)); 
		m.memref(pc+1) = (lookup_label(tokens[shift+1].data.str, pc)-m.memref(pc+2).to_int())/729;
	       	return 3; 
	} if(match_syntax(shift, LBL, PLUS, NUM, COMMA, Y, END)) { // Absolute,Y addressing (lbl+offset)
	        m.memref(pc) = m.qop(machine::AY, opcode);
		int maddr = lookup_label(tokens[shift+1].data.str, pc) + tokens[shift+3].data.num;
		tryte::int_to_word(maddr, m.memref(pc+1), m.memref(pc+2));
	       	return 3; 
	} if(match_syntax(shift, X, COMMA, Y, END)) { // XY-addressing
		m.memref(pc) = m.qop(machine::XY, opcode);
		return 1;
	}
       
	tryte high, low;
	tryte::int_to_word(pc, high, low);
	printf("\aSyntax mismatch @%.3x:%.3x\n", high.nonaryhex(), low.nonaryhex());
	getchar();

	return 0;
}

char lines[2000][80];
int linen = 0;

char asciitoternary(char a) {
	char translation[] = {  0, ' ','\n','\t','\b',   0,  0,   0,   0,   0,
			    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
			    'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
			    'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3',
			    '4', '5', '6', '7', '8', '9', '.', ',', '!', '?',
			    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
			    'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
			    'u', 'v', 'w', 'x', 'y', 'z',  0 ,   0,   0,   0,
			      0,   0,   0,   0,   0,   0, '=', '-', '*', '/',
			    '%', '<', '>',   0,   0, '(', ')', '$', '+', '#' };

	for(int c  =0; c < 100; c++) {
		if(translation[c] == a) return c;

	}
	return 0;

}

/* XXX: There is no recursion checking. Also, this assembler really doesn't
 * support inclusions. It just pretends it's all one long file. */
void readfile(FILE* file) {
	static int tier = 0;
	if(tier++ > 10) {
		printf("More than 10 inclusion tiers\n"
		       "Bailing out (most likely circular inclusion)\n");
	}
	while(linen < 2000 && fgets(lines[linen], 80, file)) {
		tokenize(lines[linen]);
	
		if(tokenn == 2 && tokens[0].type == STR && tokens[1].type == STR && strcmp(tokens[0].data.str, "INC") == 0) {
			if(verbose) printf("Including '%s' (tier %d)\n", 
					tokens[1].data.str, tier);
			FILE* inc = fopen(tokens[1].data.str, "r");
			if(inc == NULL) {
				printf("Error attempting to "
				"include %s\n", tokens[1].data.str);
				exit(EXIT_FAILURE);
			}

			readfile(inc);
			fclose(inc);
			continue;
		}
		linen++;

	}
	tier --;
}

void help(char* cmd) {
	printf("Usage: %s [-l] [-h] [-v] [-o file] file.asm\n"
		"\t-h   Show this message\n"
		"\t-v   Verbose mode\n"
		"\t-o   Set output file\n\n", cmd);
}

int main(int argc, char* argv[]) {
	int pc = tryte::word_to_int("DDD", "DDD");
	FILE* file = stdin;
	char* outfile = NULL;

	if(geteuid() == 0) {
		printf("WARNING!!! Running as root is NOT ADVISABLE !!!\n\n");
	}

	int optret;
	while((optret = getopt(argc, argv, "o:hvc")) != -1) {
		switch(optret) {
			case 'h': help(argv[0]); exit(EXIT_SUCCESS);
			case 'o': outfile = strdup(optarg); break;
			case 'v': verbose = 1; break;
			case '?': help(argv[0]); exit(EXIT_FAILURE);
			case ':': help(argv[0]); exit(EXIT_FAILURE);
		}
		
	}

	if(argc > optind) {
		file = fopen(argv[optind], "r");
		if(file == NULL) {
			printf("Unable to open \"%s\" for reading.\n",
					argv[optind]);
			exit(EXIT_FAILURE);
		}

	}

	machine m;
	
	if(verbose) printf("Reading sourcecode...\n");

	readfile(file);
	if(file != stdin) fclose(file);

	if(verbose) printf("Primary sweep...\n");

	/* one sweep to identify labels */
	shutup = 1;

	for(int line = 0; line < linen; line++) {
		tokenize(lines[line]);
		if(tokenn == 0) continue;
		int shift = 0;
		if(tokens[shift].type == LBL) {
			add_label(tokens[shift].data.str, pc);
			shift ++;
		}

		if(tokens[shift].type == STR)
			/* Set program counter */
			if(strcmp(tokens[shift].data.str, "ORG") == 0) {
				if(tokenn < 2 || tokens[shift+1].type != NUM) printf("%d: ORG with no/invalid argument!\n", line);
				pc = tokens[shift+1].data.num;
				continue;
			} else if(strcmp(tokens[shift].data.str, "DT") == 0) {
				for(int i = shift+1; i < tokenn; i++) {
					if(tokens[i].type == NUM) pc++;
					else if(tokens[i].type == STR) {
						pc += strlen(tokens[i].data.str);
					}
				}
			} else if(strcmp(tokens[shift].data.str, "DW") == 0) {
				for(int i = shift+1; i < tokenn; i++) {
					if(tokens[i].type == NUM) pc+=2;
					else if(tokens[i].type == LBL) pc+=2;
				}
				continue;
			} else if(strcmp(tokens[shift].data.str, "REST") == 0) {
				if(tokens[shift+1].type == NUM) if(shift < tokenn) { pc += tokens[shift+1].data.num; }
				continue;
			} else if(strtoopc(tokens[shift].data.str) != 0xdeadbeef) {
				pc += addop(strtoopc(tokens[shift].data.str), shift, m, pc);
				continue;
			} else {
				printf("%d: Unknown command '%s'!\n", line, tokens[shift].data.str);
			}
	}

	if(verbose) {
		printf("LABELS: \n");
		for(int i = 0; i < labeln; i++) {
			tryte low, high;
			tryte::int_to_word(labels[i].mempos, high, low);
			printf(" %.3X:%.3X (%7d)\t%s\n", high.nonaryhex(), 
					low.nonaryhex(), 
					labels[i].mempos, labels[i].label);
		}
	}

	if(verbose) printf("Secondary sweep...\n");

	shutup = 0;
	pc = tryte::word_to_int("DDD", "DDD");
	/* One sweep to assemble, with precomputed labels */
	for(int line = 0; line < linen; line++) {
		tokenize(lines[line]);
		if(tokenn == 0) continue;
		int shift = 0;
		
		if(tokens[0].type == LBL) { shift = 1; }
		if(shift == tokenn) continue;

		if(tokens[shift].type == STR) {
			/* Set program counter */
			if(strcmp(tokens[shift].data.str, "ORG") == 0) {
				if(tokenn < 2 || tokens[shift+1].type != NUM) printf("%d: ORG with no/invalid argument!\n", line);
				pc = tokens[shift+1].data.num;
				continue;
			} else if(strcmp(tokens[shift].data.str, "DT") == 0) {
				for(int i = shift+1; i < tokenn; i++) {
					if(tokens[i].type == NUM) m.memref(pc++) = tokens[i].data.num;
					else if(tokens[i].type == STR) {
						int idx;
						for(idx = 0; tokens[i].data.str[idx] != 0; idx++) {
							m.memref(pc++) = asciitoternary(tokens[i].data.str[idx]);
						}
					}
				}
				continue;
			} else if(strcmp(tokens[shift].data.str, "DW") == 0) {
				for(int i = shift+1; i < tokenn; i++) {
					if(tokens[i].type == NUM) {
						tryte::int_to_word(tokens[i].data.num, m.memref(pc), m.memref(pc+1));
						pc+=2;
					} else if(tokens[i].type == LBL) {
						tryte::int_to_word(lookup_label(tokens[i].data.str, pc), m.memref(pc), m.memref(pc+1));
					       	pc+=2;
					}
				}
			} else if(strcmp(tokens[shift].data.str, "REST") == 0) {
				if(tokenn == shift + 2 && 
						tokens[shift+1].type == NUM) { 
					int bytes = tokens[shift+1].data.num;
					while((bytes--)) {
						m.memref(pc++) = 0;
					}	
				
				} else if(tokenn == shift+3 &&
						tokens[shift+1].type == NUM &&
						tokens[shift+2].type == NUM) {

					int bytes = tokens[shift+1].data.num;
					while((bytes--)) {
						m.memref(pc++) = tokens[shift+2].data.num;
					}	
				}
				continue;
			} else if(strtoopc(tokens[shift].data.str) != 0xdeadbeef) {
				int dp = addop(strtoopc(tokens[shift].data.str), shift, m, pc);
				pc += dp;
				continue;
			} else {
				printf("%d: Unknown command '%s'!\n", line, tokens[shift].data.str);
			} 
		} else if(shift != tokenn) { 
			printf("???: %s\n", lines[line]);
		}
	} 
	if(verbose) printf("Saving... ");
	
	if(!outfile) m.save("out.ternobj");
	else m.save(outfile);

	if(verbose) printf("Done!\n");
}

