/* Tunguska, ternary virtual machine
 *
 * Copyright (C) 2007,2008 Viktor Lofgren
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "machine.h"
#include <string.h>
#include <stdio.h>
#include "tryte.h"

machine::machine() {
	init_translation_table();

	S = "444";	// Stack page is "DDC"
	SP = "DDC";

	A = X = Y = CL = 0;
	memory = new tryte[MEMSIZ];

	P = 0; /* All flags false */
	PCH = 0; PCL = 0;

	/* For sanity's sake, set a RTI interrupt vector!!! */
	memref("444", "444") = qop(IMPLICIT, RTI); 
	memref("444", "443") = "444";
	memref("444", "442") = "444";
	memref("444", "441") = 100; // Clock frequency

	autorun = true;
	trace = false;
}


char machine::translation_table[80][10];
void machine::init_translation_table() {
	/* This is a really ugly hack, but it does keep maintenance in check,
	 * and makes debugging and assembling so much easier. */
	memset(translation_table, '\0', sizeof(translation_table));
	#define STRTOOPC(X) strncpy(machine::translation_table[40+machine::X], #X, sizeof(machine::translation_table[0]));
	STRTOOPC(NOP); STRTOOPC(TAX); STRTOOPC(TAY); STRTOOPC(TXA); 
	STRTOOPC(TYA); STRTOOPC(PSH); STRTOOPC(PHP); STRTOOPC(PLL); 
	STRTOOPC(PLP); STRTOOPC(PHX); STRTOOPC(PHY); STRTOOPC(PLX);
	STRTOOPC(PLY); STRTOOPC(STA); STRTOOPC(STX); STRTOOPC(STY); 
	STRTOOPC(LDA); STRTOOPC(LDX); STRTOOPC(LDY); STRTOOPC(AND); 
	STRTOOPC(EOR); STRTOOPC(PRM); STRTOOPC(ORA); STRTOOPC(ADD);
	STRTOOPC(CMP); STRTOOPC(INC); STRTOOPC(INX); STRTOOPC(INY); 
	STRTOOPC(DEC); STRTOOPC(DEX); STRTOOPC(DEY); STRTOOPC(ASL); 
	STRTOOPC(LSR); STRTOOPC(ROL); STRTOOPC(ROR); STRTOOPC(JCC); 
	STRTOOPC(JCT); STRTOOPC(JCF); STRTOOPC(JNE); STRTOOPC(JLT); 
	STRTOOPC(JGT); STRTOOPC(JVS); STRTOOPC(JVC); STRTOOPC(CLC); 
	STRTOOPC(CLI); STRTOOPC(IVC); STRTOOPC(CLV); STRTOOPC(SEC); 
	STRTOOPC(SEI); STRTOOPC(BRK); STRTOOPC(BRK); STRTOOPC(RTI); 
	STRTOOPC(MLH); STRTOOPC(MLL); STRTOOPC(MOD); STRTOOPC(DIV); 
	STRTOOPC(JMP); STRTOOPC(JMP); STRTOOPC(JSR); STRTOOPC(RST); 
	STRTOOPC(TSH); STRTOOPC(BUT); STRTOOPC(JEQ); STRTOOPC(LAD); 
	STRTOOPC(DEBUG);
#undef STRTOOPC
}

char* machine::opcode_to_string(int opcode) {
	if(opcode < -40 || opcode > 40) {
		return translation_table[DEBUG+40];
	}
	else return translation_table[opcode+40];
}

tryte& machine::memref(int pos) {
	if(pos < -MEMSIZ/2) { pos = 0; printf("Attempt at accessing memory outside limits\n"); }
	else if(pos > MEMSIZ/2) { pos = 0; printf("Attempt at accessing memory outside limits\n"); }
	return memory[pos+MEMSIZ/2];
}

tryte& machine::memref(const tryte& high, const tryte& low) {
	int pos = tryte::word_to_int(high, low); //low.tword_to_int(high);
	return memref(pos);
}

/* Functions for loading and saving memory dumps */
/* Since most of the time these are mainly full of zeros,
 * compression can yield 100-fold or more size reduction */
void machine::load(char* filename) {
	FILE* f = fopen(filename, "r");
	if(!f) { perror("fopen"); return; }
	short data;
	for(int t = 0; t < MEMSIZ; t++) {
		fread(&data, 1, sizeof(short), f);
		memory[t] = data;
	}
	fclose(f);
}
void machine::save(char* filename) {
	FILE* f = fopen(filename, "w");
	if(f == 0) { perror(filename); return; }
	short data;
	for(int t = 0; t < MEMSIZ; t++) {
		data = memory[t].to_int();
		fwrite(&data, 1, sizeof(short), f);
	}
	fclose(f);
}

/* Process a single instruction */
void machine::instruction() {
	int ip = tryte::word_to_int(PCH, PCL);
	tryte instruction = memref(PCH, PCL);
	tryte highbits = instruction >> 4;
	instruction = (instruction << 2) >> 2;
	tryte* arg = &A;
	tryte* abs1 = &PCL,* abs2 = &PCH;
	
	/* This allows reference passing when there isn't a distinct
	 * memory address specified (absolute-X addressing for an instance) */
	static tryte cheatarg1, cheatarg2;

	if(trace) printf("%d %d\n", PCH.to_int(), PCL.to_int());

	/* Increase clock */
	CL = CL + 1;
	if((CL > memref("444", "441")) == 1) CL = 0;

	/* Address translation (a bit of a mess)*/
	switch(highbits.to_int()) {
		case ACC: arg = &A; break;
		case ABS: abs1 = &memref(ip+1);
			  abs2 = &memref(ip+2);
			  
			  arg = &memref(*abs1,*abs2); 
			break;
		case INDIRECT:
			abs1 = &memref(tryte::word_to_int(memref(ip+1), memref(ip+2)));
			abs2 = &memref(tryte::word_to_int(memref(ip+1), memref(ip+2))+1);
			arg = &memref(*abs1,*abs2); 
			break;
		case AX: 
			tryte::int_to_word(tryte::word_to_int(memref(ip+1), memref(ip+2)) + X.to_int(),
						cheatarg1, cheatarg2);
			abs1 = &cheatarg1;
			abs2 = &cheatarg2;
			arg = &memref(*abs1, *abs2);
			break;
		case AY:
			tryte::int_to_word(tryte::word_to_int(memref(ip+1), memref(ip+2)) + Y.to_int(),
						cheatarg1, cheatarg2);
			abs1 = &cheatarg1;
			abs2 = &cheatarg2;
			arg = &memref(*abs1, *abs2);
			break;
		case IMMEDIATE: arg = &memref(ip+1); break;
		case INDX: 
			tryte::int_to_word(tryte::word_to_int(memref(ip+1), memref(ip+2)) + X.to_int(),
					cheatarg1, cheatarg2);
			abs1 = &memref(cheatarg1, cheatarg2);
			abs2 = &memref(tryte::word_to_int(cheatarg1, cheatarg2) + 1);
			arg = &memref(*abs1, *abs2);
			break;
		case INDY: 
			tryte::int_to_word(tryte::word_to_int(memref(ip+1), memref(ip+2)) + Y.to_int(),
					cheatarg1, cheatarg2);
			abs1 = &memref(cheatarg1, cheatarg2);
			abs2 = &memref(tryte::word_to_int(cheatarg1, cheatarg2) + 1);
			arg = &memref(*abs1, *abs2);

			arg = &memref(tryte::word_to_int(memref(ip+1), memref(ip+2)) + Y.to_int());
			break;
		case XY:
			abs1 = &X;
			abs2 = &Y;
			arg = &memref(*abs1, *abs2);
			break;
	}
	switch(highbits.to_int()) {
		case XY:
		case ACC:
			if(((PCL + 1) < PCL) == 1) PCH = PCH + 1;
			PCL += 1; 
			break;
		case IMMEDIATE:
			if(((PCL + 2) < PCL) == 1) PCH = PCH + 1;
			PCL += 2; 
			break;
		case AX:
		case AY:
		case INDX:
		case INDY:
		case ABS:
		case INDIRECT:
			if(((PCL + 3) < PCL) == 1) PCH = PCH + 1;
			PCL += 3;
			break;
	}

	switch(instruction.to_int()) {
		case NOP: nop(); break;
		case TAX: tax(); break;
		case TAY: tay(); break;
		case TXA: txa(); break;
		case TYA: tya(); break;
		case PSH: psh(arg); break;
		case PHP: php(); break;
		case PLL: pll(arg); break;
		case PLP: plp(); break;
		case PHX: phx(); break;
		case PHY: phy(); break;
		case PLX: plx(); break;
		case PLY: ply(); break;
		case STA: sta(arg); break;
		case STX: stx(arg); break;
		case STY: sty(arg); break;
		case LDA: lda(arg); break;
		case LDX: ldx(arg); break;
		case LDY: ldy(arg); break;
		case AND: _and(arg); break;
		case EOR: eor(arg); break;
		case PRM: prm(arg); break;
		case ORA: ora(arg); break;
		case BIT: bit(arg); break;
		case ADD: add(arg); break;
		case CMP: cmp(arg); break;
		case INC: inc(arg); break;
		case INX: inx(); break;
		case INY: iny(); break;
		case DEC: dec(arg); break;
		case DEX: dex(); break;
		case DEY: dey(); break;
		case ASL: asl(arg); break;
		case LSR: lsr(arg); break;
		case ROL: rol(arg); break;
		case ROR: ror(arg); break;
		case RST: rst(); break;

		case JCC: jcc(abs1, abs2); break;
	       	case JCT: jct(abs1, abs2); break;
	       	case JCF: jcf(abs1, abs2); break;
	       	case JEQ: jeq(abs1, abs2); break;
	       	case JNE: jne(abs1, abs2); break; 
		case JGT: jgt(abs1, abs2); break; 
		case JLT: jlt(abs1, abs2); break; 
		case JVC: jvc(abs1, abs2); break; 
		case JVS: jvs(abs1, abs2); break;

		case CLC: clc(); break;
		case IVC: ivc(); break; /* invert carry */
		case CLI: cli(); break;
		case CLV: clv(); break;

		case SEC: sec(); break;
		case SEI: sei(); break;

		case BRK: soft_brk(); break;
		case RTI: rti(); break;
		case MLH: mlh(arg); break;
		case MLL: mll(arg); break;
		case MOD: mod(arg); break;
		case DIV: div(arg); break;
		case JMP: jmp(abs1, abs2); break;
		case LAD: lad(abs1, abs2); break;
		case JSR: jsr(abs1, abs2); break;
		case DEBUG: debug(); break;

		default: printf("%d: unknown opcode %d\n", ip, instruction.to_int());
	}

}

void machine::pflag(const tryte& t) {
	if(t.to_int() == 0) P[G] = trit::TRMU;
	else if(t.to_int() > 0) P[G] = trit::TRTRUE;
	else P[G] = trit::TRFALSE;
	P[PR] = t.parity();

	P[C] = t.carry;
}

void machine::debug() {
	tryte instruction = memref(PCH, PCL);
	instruction = (instruction << 2) >> 2;
	printf("---DEBUG---\n"
		"PC:%d:%d(%s)\tX:%d\tY:%d\tACC:%d\n", PCH.to_int(), PCL.to_int(), opcode_to_string(instruction.to_int()), X.to_int(), Y.to_int(), A.to_int());
	printf("P: \tP%d\t V%d\t B%d\t I%d\t G%d\t C%d\n", P[PR].to_int(), P[V].to_int(), P[B].to_int(), P[I].to_int(), P[G].to_int(), P[C].to_int());
	printf("S: %d\n", S.to_int());
	printf("CL: %d\n", CL.to_int());
	fflush(NULL);
}

/* Truck load of nothing */
void machine::nop() { }

/* Loaders */
void machine::lda(tryte *t) { A = *t; pflag(A); }
void machine::ldx(tryte *t) { X = *t; pflag(X); }
void machine::ldy(tryte *t) { Y = *t; pflag(Y); }

/* Storers */
void machine::sta(tryte *t) { *t = A; pflag(*t); }
void machine::stx(tryte *t) { *t = X; pflag(*t); }
void machine::sty(tryte *t) { *t = Y; pflag(*t); }

/* Transferrers */
void machine::tax() { X = A; pflag(X); }
void machine::tay() { Y = A; pflag(Y); }
void machine::txa() { A = X; pflag(A); }
void machine::tya() { A = Y; pflag(A); }
void machine::tsx() { X = S; pflag(X); }
void machine::txs() { S = X; pflag(S); }

/* Pushers */
void machine::psh(tryte* t) { S += -1; memref(SP,S)=*t; pflag(*t); }
void machine::phx() { S += -1; memref(SP,S)=X; pflag(X); }
void machine::phy() { S += -1; memref(SP,S)=Y; pflag(Y); }
void machine::php() { S += -1; memref(SP,S)=P; }

/* Pullers */
void machine::pll(tryte* t) { *t = memref(SP,S); S += 1; pflag(*t); }
void machine::plx() { X = memref(SP,S); S += 1; pflag(X); }
void machine::ply() { Y = memref(SP,S); S += 1; pflag(Y); }
void machine::plp() { P = memref(SP,S); S += 1; }

/* Logical operations */
void machine::_and(tryte* t) { A = A & (*t); pflag(A); }
void machine::eor(tryte* t) { A = A ^ (*t); pflag(A); }
void machine::prm(tryte* t) { A.permute_this(*t); pflag(A); }
void machine::but(tryte* t) { A.but_this(*t); pflag(A); }
void machine::tsh(tryte* t) { A.shift_this(*t); pflag(A); }
void machine::ora(tryte* t) { A = A | *t; pflag(A); }
void machine::bit(tryte* t) { pflag(A & *t); }

/* Arithmetics */
void machine::add(tryte* t) { 
	if((t->to_int() + P[C].to_int()) > 0 && (A + *t + P[C].to_int()) < A) 
		P[V] = trit::TRTRUE;
	else if((t->to_int() + P[C].to_int()) < 0 && (A + *t + P[C].to_int()) > A)
		P[V] = trit::TRFALSE;
	else P[V] = trit::TRMU;
	A = A.to_int() + t->to_int() + P[C].to_int(); 
	pflag(A); 
}
void machine::cmp(tryte* t) {
	if(t > 0 && (A + *t) < A) 
		P[V] = trit::TRFALSE;
	else if(t < 0 && (A + *t) > A)
		P[V] = trit::TRTRUE;
	else P[V] = trit::TRMU;
	pflag(A + -(t->to_int())); 
}

void machine::inc(tryte* t) { *t = *t + 1; pflag(*t); 
	P[V] = (*t == -364)>0 ? trit::TRTRUE : trit::TRMU; }
void machine::inx() { X = X + 1; pflag(X); 
	P[V] = (X == -364)>0 ? trit::TRTRUE : trit::TRMU; }
void machine::iny() { Y = Y + 1; pflag(Y); 
	P[V] = (Y  == -364)>0 ? trit::TRTRUE : trit::TRMU; }
void machine::dec(tryte* t) { *t = *t + -1; pflag(*t);
	P[V] = (*t == 364)>0 ? trit::TRFALSE: trit::TRMU; }
void machine::dex() { X = X + -1; pflag(X);
	P[V] = (X == 364)>0 ? trit::TRFALSE: trit::TRMU; }
void machine::dey() { Y = Y + -1; pflag(Y);
	P[V] = (Y == 364)>0 ? trit::TRFALSE: trit::TRMU; }

/* Shifts and rolls */
void machine::asl(tryte* t) { t->carry = (*t)[5]; *t = *t << 1; pflag(*t);}
void machine::lsr(tryte* t) { t->carry = (*t)[0]; *t = *t >> 1; pflag(*t); }
void machine::ror(tryte* t) { *t = (*t >> 1) + (*t << 4) >> 4; pflag(*t); }
void machine::rol(tryte* t) { *t = (*t << 1) + (*t >> 4) << 4; pflag(*t); }

/* Jumps */
void machine::jmp(const tryte* high, const tryte* low) { 
	PCH = *high; PCL = *low; 

}
void machine::jsr(const tryte* high, const tryte* low) {
	S += -1; memref(SP, S) = PCH;
	S += -1; memref(SP, S) = PCL;
	PCH = *high; PCL = *low; 
}
void machine::rst() { 
	PCL = memref(SP, S); S += 1;
	PCH = memref(SP, S); S += 1;
}

void machine::jct(const tryte* t, const tryte* t2) {
	if((P[C]).to_int() == 1) jmp(t, t2);
}
void machine::jcf(const tryte* t, const tryte* t2) {
	if((P[C]).to_int() == -1) jmp(t, t2);
}
void machine::jcc(const tryte* t, const tryte* t2) {
	if((P[C]).to_int() == 0) jmp(t, t2);
}
void machine::jeq(const tryte* t, const tryte* t2) {
	if((P[G]).to_int() == 0) jmp(t, t2);
}
void machine::jne(const tryte* t, const tryte* t2) {
	if((P[G]).to_int() != 0) jmp(t, t2);
}
void machine::jgt(const tryte* t, const tryte* t2) {
	if((P[G]).to_int() == 1) jmp(t, t2);
}
void machine::jlt(const tryte* t, const tryte* t2) {
	if((P[G]).to_int() == -1) jmp(t, t2);
}
void machine::jvs(const tryte* t, const tryte* t2) {
	if((P[V]).to_int() == 1) jmp(t, t2);
}
void machine::jvc(const tryte* t, const tryte* t2) {
	if((P[V]).to_int() <= 0) jmp(t, t2);
}

/* Flag manipulation */
void machine::clc() { P[C] = trit::TRMU; }
void machine::cli() { P[I] = trit::TRMU; }
void machine::ivc() { P[C] = !P[C]; }
void machine::clv() { P[V] = trit::TRMU; }
void machine::sec() { P[C] = trit::TRTRUE; }
void machine::sei() { P[I] = trit::TRTRUE; }

/* More arithemetics */
void machine::mlh(tryte* t) { A = (*t) , A; pflag(A); }
void machine::mll(tryte* t) { A = (*t) * A; pflag(A); }
void machine::div(tryte* t) { A = A / (*t); pflag(A); }
void machine::mod(tryte* t) { A = A % (*t); pflag(A); }

/* Load ADdress */
void machine::lad(const tryte* t, const tryte* t2) {
	X = *t;
	Y = *t2;
}

void machine::soft_brk() {
	if(P[I].to_int() == 0 && P[B].to_int() == 0) {
		memref("DDD", "DDD") = IRQ_USER;
		brk();
	} /* STUB */
}

/* Break (hardware or software) */
void machine::brk() {
	S += -1; memref(SP,S) = PCH;
	S += -1; memref(SP,S) = PCL;
	S += -1; memref(SP,S) = P;

	S += -1; memref(SP,S) = X;
	S += -1; memref(SP,S) = Y;
	S += -1; memref(SP,S) = A;

	P[B]= trit::TRTRUE;
	P[I]= trit::TRTRUE;

	PCL = memref(364, 363);
	PCH = memref(364, 362);
}

/* Return from Interrupt */
void machine::rti() {
	if((P[B].to_int()) <= 0) printf("RTI outside BRK\n");

	A   = memref(SP,S); S += 1;
	Y   = memref(SP,S); S += 1;
	X   = memref(SP,S); S += 1;
	P   = memref(SP,S); S += 1;
	PCL = memref(SP,S); S += 1;
	PCH = memref(SP,S); S += 1;
}
