/* Tunguska, ternary virtual machine
 *
 * Copyright (C) 2007,2008 Viktor Lofgren
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "trit.h"
#include <stdlib.h>
#include <stdio.h>

trit& trit::operator=(const trit& t) {
	data = t.data;
	return *this;
}

trit& trit::operator=(int i) {
	data = i;
	return *this;
}

trit trit::operator!() const {
	switch(data) {
		case TRFALSE: return trit(TRTRUE);
		case TRMU: return trit(TRMU);
		case TRTRUE: return trit(TRFALSE);
		default: return trit(TRMU);
	}
}

trit trit::operator&(const trit& t) const {
	if(t.data == TRTRUE && data == TRTRUE) return trit(TRTRUE);
	if((t.data == TRMU || data == TRMU)) return trit(TRMU);
	return trit(TRFALSE);
}

trit trit::operator|(const trit& t) const {
	if(t.data == TRTRUE || data == TRTRUE) return trit(TRTRUE);
	if(t.data == TRMU || data == TRMU) return trit(TRMU);
	return trit(TRFALSE);
}

trit trit::operator^(const trit& t) const {
	return -1 * t.data * data;
}

trit& trit::operator|=(const trit& t) {
	data = (*this | t).data;
	return *this;
}

trit& trit::operator&=(const trit& t) {
	data = (*this & t).data;
	return *this;
}

trit& trit::operator^=(const trit& t) {
	data = (*this ^ t).data;
	return *this;
}

int trit::operator==(const trit& t) const {
	if(t.data == this->data) return TRFALSE;
	else return TRTRUE;
}

int trit::operator!=(const trit& t) const {
	return -1 * (*this == t);
}

/* Rotation operator, it shifts the value in the direction t
 *
 * It is sort-of related to binary NOT operator, in the sense that
 * the same permutation in a binary system would result in NOT. 
 *
 * It has some useful properties, 
 * associative and commutative, and 3-cyclic
 *
 * A perm B perm NOT B = A perm B perm B perm B = A
 *
 * Truth table:
 * A | T | A > B 
 * --+---+---------
 * + | + |  -       
 * 0 | + |  +
 * - | + |  0
 * + | 0 |  +
 * 0 | 0 |  0
 * - | 0 |  -
 * + | - |  0
 * 0 | - |  -
 * - | - |  +
 *
 * */
trit trit::operator>(const trit& t) const { 
	return trit((1 + t.data + data) % 3 - 1);
}

/* Same as operator>, only it doesn't roll over 
 *
 * A | T | A > B 
 * --+---+---------
 * + | + |  +       
 * 0 | + |  +
 * - | + |  0
 * + | 0 |  +
 * 0 | 0 |  0
 * - | 0 |  -
 * + | - |  0
 * 0 | - |  -
 * - | - |  -
 *
 * */
trit trit::operator>>(const trit& t) const { 
	if(t.data == 1) {
		return trit(abs((1 + t.data + data) % 3 - 1));
	} else if(t.data == -1) {
	return trit(-abs((1 + t.data + data) % 3 - 1));
	} else return trit(*this);
}

/* Ternary BUT operator, complement to AND and OR
 *
 * Truth table:
 *
 * A | B | A BUT B
 * --+---+--------
 * - | - | -
 * 0 | - | 0
 * + | - | -
 * - | 0 | 0
 * 0 | 0 | 0
 * + | 0 | 0
 * - | + | -
 * 0 | + | 0
 * + | + | +
 *
 *
 */
trit trit::but(const trit& t) const {
	if(t.data == TRMU || data == TRMU) return trit(TRMU);
	if(t.data == TRFALSE|| data == TRFALSE) return trit(TRFALSE);
	return trit(TRTRUE);
}

trit& trit::but_this(const trit& t) {
	if(t.data == TRMU || data == TRMU) data = TRMU;
	else if(t.data == TRFALSE|| data == TRFALSE) data = TRFALSE;
	else data = TRTRUE;

	return *this;
}

