/* Tunguska, ternary virtual machine
 *
 * Copyright (C) 2007,2008 Viktor Lofgren
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "trit.h"
#include "tryte.h"
#include "machine.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>

tryte::tryte() {
	for(int i = 0; i < 6; i++) {
		trits[i] = trit::TRMU;
	}
	carry = trit::TRMU;
}

tryte::tryte(const tryte& t) {
	for(int i = 0; i < 6; i++) {
		this->trits[i] = t[i];
	}
	carry = t.get_carry();

}
tryte::tryte(trit* trits) {
	for(int i = 0; i < 6; i++) {
		this->trits[i] = trits[i];
	}
}

/* From balanced base9 */
tryte::tryte(const char* s) {
	int len = strlen(s);
	char* translation = "DCBA01234";
	if(len != 3) printf("Malformed balanced nonary numeral length\n");
	int power = 81; int num; int val = 0;
	for(num = 0; num < 3; num++) {
		char* idx = index(translation, s[num]);
		if(idx == NULL) {
		       	printf("Error in balanced nonary numeral conversion\n");
			break;
		} else {
			val += power * (unsigned int)(idx - translation - 4);
			power /= 9;
		}
	}

	*this = tryte( val );
	
}

/* This is one ugly algorithm */
tryte::tryte(int i) {
	register char t; 
	register char invert = 0; 
	register int power = 177147;

	if(i > 364) carry = trit::TRTRUE;
	else if(i < -364) carry = trit::TRFALSE;
	else carry = trit::TRMU;

	for(t = 11; t >= 0; --t) {
		if(i < 0) { invert = !invert; i *= -1; }
		if((i - power + (power-1)/2) >= 0) {
			if(t <= 5) trits[5 - (int)t] = ((invert) ? trit::TRFALSE : trit::TRTRUE);
			i -= power;
		} 
		power /= 3;
	}
}

//int tryte :: tword_to_int(const tryte& high) const {
//	return to_int() + 3 * 243 * high.to_int();
//}

tryte tryte :: operator~() const {
	trit tritset[6];
	for(int i = 0; i < 6; i++) tritset[i] = !trits[i];
	return tryte(tritset);
}

tryte tryte :: operator&(const tryte a) const {
	trit tritset[6];
	for(int i = 0; i < 6; i++) tritset[i] = (a.trits[i]) & (trits[i]);
	return tryte(tritset);
}

tryte tryte :: operator|(const tryte a) const {
	trit tritset[6];
	for(int i = 0; i < 6; i++) tritset[i] = (a.trits[i]) | (trits[i]);
	return tryte(tritset);
}

tryte tryte :: operator^(const tryte a) const {
	trit tritset[6];
	for(int i = 0; i < 6; i++) tritset[i] = (a.trits[i]) ^ (trits[i]);
	return tryte(tritset);
}

tryte& tryte :: operator&=(tryte &a) {
	(*this) = ((*this) & a);
	return *this;
}

tryte& tryte :: operator|=(tryte &a) {
	(*this) = ((*this) | a);
	return *this;
}

int tryte::operator==(const tryte &a) const {
	if(to_int() == a.to_int()) return trit::TRTRUE; 
	return trit::TRFALSE;
}

int tryte::operator!=(const tryte &a) const { return -1 * (*this == a); }

int tryte::operator>(const tryte& a) const {
	if(to_int() > a.to_int()) return trit::TRTRUE;
	else if(to_int() == a.to_int()) return trit::TRMU;
	else return trit::TRFALSE;
}

int tryte::operator<(const tryte& a) const {
	return -1 * ((*this) > a);
}

tryte& tryte :: operator=(const tryte &a) {
	int i = 0;
	for(i = 0; i < 6; i++) trits[i] = a[i];
	carry = a.get_carry();
	return *this;
}
tryte& tryte :: operator=(const int v) {
	operator=((tryte)v);
	return *this;
}
tryte tryte :: operator<<(int s) const {
	tryte t;
	for(int i = 0; i < 6-s; i++) { t[i] = (*this)[i+s]; }
	return t;
}
tryte tryte :: operator>>(int s) const {
	tryte t;
	for(int i = 0; i < 6-s; i++) { t[i+s] = (*this)[i]; }
	return t;
}

tryte tryte::operator+(const tryte& a) const {
	return to_int() + a.to_int();
}
tryte& tryte::operator+=(const tryte& a) {
	*this = tryte(to_int() + a.to_int());
	return *this;
}
tryte& tryte::operator+=(const int v) {
	*this = tryte(to_int() + v);
	return *this;
}

tryte tryte::operator+(const int v) const {
	return to_int() + v;
}

/* Low tryte in multiplication */
tryte tryte::operator*(const tryte& a) const {
	return tryte(to_int() * a.to_int());
}

/* High tryte in multiplication */
tryte tryte::operator,(const tryte& a) const {
	int v = to_int() * a.to_int();
	return (v - tryte(v).to_int()) / 729;
}

tryte tryte::operator%(const tryte& a) const {
	if(a.to_int() == 0) return 0;
	return to_int() % a.to_int();
}

tryte tryte::operator%(const int v) const {
	if(v == 0) return 0;
	return to_int() % v;
}

tryte tryte::operator/(const tryte& a) const {
	if(a.to_int() == 0) return 0;
	return to_int() / a.to_int();
}

tryte tryte::operator/(const int v) const {
	if(v == 0) return 0;
	return to_int() / v;
}

tryte tryte::permute(const tryte& t) const {
	tryte ret = t;
	for(int i = 0; i < 6; i++) {
		ret.trits[i] = ret.trits[i] > trits[i];
	}
	return ret;
}

tryte& tryte::permute_this(const tryte& t) {
	for(int i = 0; i < 6; i++) {
		trits[i] = trits[i] > t.trits[i];
	}
	return *this;
}

tryte tryte::shift(const tryte& t) const {
	tryte ret = t;
	for(int i = 0; i < 6; i++) {
		ret.trits[i] = ret.trits[i] >> trits[i];
	}
	return ret;
}

tryte& tryte::shift_this(const tryte& t) {
	for(int i = 0; i < 6; i++) {
		trits[i] = trits[i] >> t.trits[i];
	}
	return *this;
}

tryte tryte::but(const tryte& t) const {
	tryte ret = t;
	for(int i = 0; i < 6; i++) {
		ret.trits[i] = ret.trits[i] >> trits[i];
	}
	return ret;
}

tryte& tryte::but_this(const tryte& t) {
	for(int i = 0; i < 6; i++) {
		trits[i].but_this(t.trits[i]);
	}
	return *this;
}

trit tryte::parity() const {
	int sum = (trits[5].to_int() + trits[4].to_int() + trits[3].to_int() +
		   trits[2].to_int() + trits[1].to_int() + trits[0].to_int()); 
	if(sum == 0) return 0;
	return sum / abs(sum);
}
